#!/usr/bin/perl -w
# Compares bootstrap and length for splits that match
use strict;
use lib "$ENV{HOME}/Genomics/Perl/modules";
use MOTree;
use Getopt::Long;

{
    my $usage = "CompareTree.pl -tree treeFile1 -versus treeFile2 [-debug 1]\n"
	. "   Compares bootstrap and length for splits in tree1 that are found\n"
	. "   in tree2. If a split in tree1 is absent from tree2, has no\n"
	. "   values for tree2\n"
	. "   Warning: this is designed for unrooted trees in which every internal node\n"
	. "   has more than one child. For rooted trees (i.e. the root has two children),\n"
	. "   the splits around the root's children may be reported more than once.\n"
	. "   This could also lead to small errors in the reported Frac of splits that match\n";
    
    my $file1 = undef;
    my $file2 = undef;
    my $debug = 0;
    die $usage unless GetOptions('tree=s' => \$file1,
				 'versus=s' => \$file2,
				 'debug=i' => \$debug)
	&& defined $file1 
	&& defined $file2
	&& @ARGV == 0;
    open(FILE1,"<",$file1) || die "Cannot read $file1";
    open(FILE2,"<",$file2) || die "Cannot read $file2";

    my $nTree = 0;
    while(1) {
	my $tree1 = MOTree::new(fh => \*FILE1);
	if (!defined $tree1) {
	    die "Cannot read tree from $file1" if $nTree == 0;
	    last;
	    next;
	}
	my $tree2 = MOTree::new(fh => \*FILE2);
	$nTree++;
	die "Cannot read tree $nTree from $file2" unless defined $tree2;

	# remove leading and trailing underscores, created as artifacts of some tree conversions e.g. Rose
	foreach my $node (@{ $tree1->depthfirst() }) {
	    if ($tree1->is_Leaf($node)) {
		my $id = $tree1->id($node);
		$id =~ s/^_+//;
		$id =~ s/_+$//;
		$tree1->set_id($node,$id);
	    }
	}
	foreach my $node (@{ $tree2->depthfirst() }) {
	    if ($tree2->is_Leaf($node)) {
		my $id = $tree2->id($node);
		$id =~ s/^_+//;
		$id =~ s/_+$//;
		$tree2->set_id($node,$id);
	    }
	}

	my $match = $tree1->Compare($tree2);
	if ($nTree==1) {
	    print join("\t", qw{Len1 Boot1 Len2 Boot2 NDesc LeafA LeafB Node1 Node2})."\n";
	}
	my $maxlenDiff = undef;
	my $maxbootDiff = undef;
	my $nDesc1 = $tree1->countAllLeaves(); # node => number of leaves beneath it
	my $totLeaf1 = $nDesc1->{$tree1->get_root_node()};
	my $nFound = 0;
	my $nInternal = 0;
	foreach my $node1 (@{ $tree1->depthfirst() }) {
	    next if $nDesc1->{$node1} <= 1 || $nDesc1->{$node1} >= $totLeaf1 - 1;
	    $nInternal++;

	    my $len1 = $tree1->branch_length($node1);
	    my $boot1 = $tree1->id($node1);

	    my $match1 = $match->{$node1};
	    my ($node2,$len2,$boot2);
	    if (defined $match1) {
		$nFound++;
		$node2 = $match1->[0];
		$len2 = $tree2->branch_length($node2);
		$boot2 = $tree2->id($node2);
	    }
	    if (defined $len1 && defined $len2 && $len1 ne "" && $len2 ne "") {
		my $diff = abs($len1-$len2);
		$maxlenDiff = $diff if !defined $maxlenDiff || $diff > $maxlenDiff;
	    }
	    if (defined $boot1 && defined $boot2 && $boot1 ne "" && $boot2 ne "") {
		my $diff = abs($boot1-$boot2);
		$maxbootDiff = $diff if !defined $maxbootDiff || $diff > $maxbootDiff;
	    }
	    my @children1 = $tree1->children($node1);
	    my $leafA = $children1[0];
	    while (! $tree1->is_Leaf($leafA)) { my @l = $tree1->children($leafA); $leafA = $l[0]; };
	    my $leafB = $children1[-1];
	    while (! $tree1->is_Leaf($leafB)) { my @l = $tree1->children($leafB); $leafB = $l[-1]; };
	    my @fields = map { defined $_ ? $_ : "" } ($len1, $boot1,
						       $len2, $boot2,
						       $nDesc1->{$node1},
						       $tree1->id($leafA), $tree1->id($leafB),
						       $node1, $node2);

	    print join("\t", @fields)."\n";
	}
	print STDERR join("\t","Splits","Found",$nFound,"Total",$nInternal,
			  "Frac",$nInternal > 0 ? sprintf("%.4g", $nFound/$nInternal) : "",
			  "MaxLnDf",defined $maxlenDiff ? sprintf("%.3g",$maxlenDiff) : "",
			  "MaxBtDf",defined $maxbootDiff? sprintf("%.3g",$maxbootDiff) : ""
			  )."\n";
    }
}

